import functools
import typing

import attrs
import naff


@attrs.define(
    eq=False, order=False, hash=False, kw_only=False, on_setattr=attrs.setters.NO_OP
)
class FasterEmbed(naff.Embed):
    @functools.lru_cache(maxsize=32)
    def __len__(self):
        # admittedly, the base lib's len is already pretty fast -
        # this doesn't speed it up a whole lot
        # the biggest speedups come from walrus operators
        # and the fact that the result of this operation is cached
        total: int = 0

        if title := self.title:
            total += len(title)
        if description := self.description:
            total += len(description)
        if footer := self.footer:
            total += len(footer)
        if author := self.author:
            total += len(author)

        total += sum(map(len, self.fields))
        return total

    @functools.lru_cache(maxsize=32)
    def __bool__(self: naff.Embed) -> bool:
        # not not is faster than any
        return not not (
            self.title
            or self.description
            or self.fields
            or self.author
            or self.thumbnail
            or self.footer
            or self.image
            or self.video
        )

    # the cache will keep invalid len/bool results if we don't clear it
    # on attribute or field change, so we make sure to clear it manually
    # that being said, using fields.append would be a problem... not like
    # there's much that can be done there

    def _invalidate_cache(self):
        self.__len__.cache_clear()
        self.__bool__.cache_clear()

    def add_field(self, name: str, value: typing.Any, inline: bool = False) -> None:
        self._invalidate_cache()
        return super().add_field(name, value, inline)

    def add_fields(self, *fields: naff.EmbedField | str | dict) -> None:
        self._invalidate_cache()
        return super().add_fields(*fields)

    def __setattr__(self, __name: str, __value: typing.Any) -> None:
        self._invalidate_cache()
        return super().__setattr__(__name, __value)


def patch():
    # why correctly monkeypatch when overriding init does the trick?
    naff.Embed.__init__ = FasterEmbed.__init__



import functools
import inspect
import typing

import attrs
import naff

from ..utils import filter_extras
from ..utils import is_optional


@attrs.define()
class BaseCommandParameter:
    name: str = attrs.field()
    type: typing.Any = attrs.field()
    kind: inspect._ParameterKind = attrs.field()
    default: typing.Any = attrs.field(default=naff.MISSING)
    converter: typing.Optional[typing.Callable] = attrs.field(default=None)


def _get_from_annotated(annotated: typing.Annotated):
    args = typing.get_args(annotated)
    return next((a for a in args if isinstance(a, naff.Converter)), None)


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class BetterBaseCommand(naff.BaseCommand):
    parameters: dict[str, BaseCommandParameter] = attrs.field(
        repr=False, factory=dict, metadata=naff.utils.no_export_meta
    )
    _uses_arg: bool = attrs.field(
        repr=False, default=False, metadata=naff.utils.no_export_meta
    )

    def __attrs_post_init__(self):
        self.parameters = {}
        self._uses_arg = False

        naff.BaseCommand.__attrs_post_init__(self)

        if self.callback is not None and not self.parameters:
            if "." in self.callback.__qualname__:
                callback = functools.partial(self.callback, None, None)
            else:
                callback = functools.partial(self.callback, None)

            sig_parameters = inspect.signature(callback).parameters

            for param in sig_parameters.values():
                if param.kind == inspect._ParameterKind.VAR_POSITIONAL:
                    self._uses_arg = True
                    continue

                if param.kind == inspect._ParameterKind.VAR_KEYWORD:
                    self._uses_arg = False
                    continue

                our_param = BaseCommandParameter(
                    param.name, param.annotation, param.kind
                )
                our_param.default = (
                    param.default
                    if param.default is not inspect._empty
                    else naff.MISSING
                )

                if param.annotation is not inspect._empty:
                    anno = param.annotation
                    converter = None

                    if is_optional(anno):
                        # base commands support annotations in a different way,
                        # so we don't want to filter them out like how tansy does
                        # however, we do want to get the non-optional stuff out of
                        # this anno, so filter_extras is still useful
                        anno = filter_extras(anno)

                    if isinstance(anno, naff.Converter):
                        converter = anno
                    elif typing.get_origin(anno) == typing.Annotated:
                        converter = _get_from_annotated(anno)

                    if converter:
                        our_param.converter = self._get_converter_function(
                            converter, our_param.name
                        )

                self.parameters[param.name] = our_param

    async def call_callback(self, callback: typing.Callable, ctx: naff.Context) -> None:
        if not self.parameters:
            # to keep compat with normal BaseCommand
            return await self.call_with_binding(callback, ctx)

        kwargs_copy = ctx.kwargs.copy()

        new_args = []
        new_kwargs = {}

        for name in ctx.kwargs.keys():
            value = kwargs_copy.pop(name)
            param = self.parameters[name]

            if converter := param.converter:
                value = await naff.utils.maybe_coroutine(converter, ctx, value)

            if param.kind == inspect.Parameter.POSITIONAL_ONLY:
                new_args.append(value)
            else:
                new_kwargs[name] = value

        if kwargs_copy:
            if self._uses_arg:
                new_args.extend(kwargs_copy.values())
            else:
                new_kwargs |= kwargs_copy

        return await self.call_with_binding(callback, ctx, *new_args, **new_kwargs)


def interaction_cmd_post_init(self: naff.InteractionCommand):
    if self.callback is not None and hasattr(self.callback, "auto_defer"):
        self.auto_defer = self.callback.auto_defer

    BetterBaseCommand.__attrs_post_init__(self)


def patch():
    naff.InteractionCommand.__attrs_post_init__ = interaction_cmd_post_init
    naff.InteractionCommand.call_callback = BetterBaseCommand.call_callback



import contextlib

from . import embed
from . import interaction_command
from . import orjson_serialize

__all__ = ("install_naff_speedups",)

HAS_ORJSON = False

with contextlib.suppress(ImportError):
    import orjson

    HAS_ORJSON = True


def install_naff_speedups(
    *,
    embeds: bool = True,
    interaction_commands: bool = True,
    use_orjson_for_serialization: bool = HAS_ORJSON,
) -> None:
    """
    Monkeypatches/installs various speedups for NAFF.

    These usually are unstable, and some may change the behavior of NAFF in unexpected way -
    while common use-case scenarios are expected to act mostly normally, Tansy has no qualms
    about rewriting internals for speed gains.

    This should be run as soon as possible if you wish to use it. Ideally, it should be
    one of the first things you do in your bot runner file, right after importing everything.
    This allows the monkeypatches to work as expected.

    If any of these patches are installed, report behavior related to them to Tansy, not NAFF.

    Args:
        embeds (bool): Speeds up length getting and validating checking in embeds,
        possibly by a significant margin if those operations are run repeatedly.
        Note: directing appending to the `fields` attribute in Embeds will cause unexpected
        behavior - it is suggested you use the `add_field(s)` functions instead while using this.

        interaction_commands (bool): Speeds up interaction commands by pre-processing
        parameters in a callback, similar to what Tansy and prefixed commands do. This saves
        from having to re-analyze the command every time it is called and adds in proper
        Annotated support (as a side effect), but breaks certain (niche) parts of commands.

        use_orjson_for_serialization (bool): By default, aiohttp, and so NAFF, use json for
        serializing payloads (when they are dicts, which is quite often). orjson is a faster
        alternative to the built-in json module, and is already used for deserialization if
        it is detected to be installed, but not for serialization. Tansy makes it so orjson
        is also used for serialization, though the speed difference is likely to be
        negligible for most operations.
        By default, this is only enabled if orjson is installed.
    """

    if embeds:
        embed.patch()
    if interaction_commands:
        interaction_command.patch()
    if use_orjson_for_serialization:
        orjson_serialize.patch()



import typing

from naff.api.http.http_client import HTTPClient

OrJsonHTTP: typing.Optional[type[HTTPClient]] = None

try:
    import orjson

    def _orjson_dumps_handler(obj):
        return orjson.dumps(obj).decode("utf-9")

    class _OrJsonHTTP(HTTPClient):
        async def login(self, token: str):
            to_return = super().login(token)
            # private? more like public
            self._HTTPClient__session._json_serialize = _orjson_dumps_handler
            return to_return

    OrJsonHTTP = _OrJsonHTTP

except ImportError:
    pass


def patch():
    if not OrJsonHTTP:
        raise ImportError("orjson is not installed.")

    HTTPClient.__init__ = OrJsonHTTP.__init__



import itertools
import types
import typing

import naff

REVERSE_CHANNEL_MAPPING = {v: k for k, v in naff.TYPE_CHANNEL_MAPPING.items()}

UNION_TYPES = {typing.Union, types.UnionType}

USER_TYPES = (naff.BaseUser, naff.User, naff.Member)
USER_PRODUCT = itertools.product(USER_TYPES, USER_TYPES, USER_TYPES)

MENTIONABLE_UNIONS = frozenset(
    typing.Union[naff.Role, i, j, k] for i, j, k in USER_PRODUCT
)
USER_UNIONS = frozenset(typing.Union[i, j, k] for i, j, k in USER_PRODUCT)


def is_union(anno: typing.Any):
    return typing.get_origin(anno) in UNION_TYPES


def get_from_anno_type(anno: typing.Annotated) -> typing.Any:
    """
    Handles dealing with Annotated annotations, getting their (first) type annotation.
    This allows correct type hinting with, say, Converters, for example.
    """
    # this is treated how it usually is during runtime
    # the first argument is ignored and the rest is treated as is

    return typing.get_args(anno)[1]


def issubclass_failsafe(
    arg: typing.Any, cls: typing.Type | typing.Tuple[typing.Type]
) -> bool:
    try:
        return issubclass(arg, cls)
    except TypeError:
        return False


def is_optional(anno: typing.Any):
    return is_union(anno) and types.NoneType in typing.get_args(anno)


def filter_extras(t: naff.OptionTypes | type):
    if typing.get_origin(t) == typing.Annotated:
        t = get_from_anno_type(t)

    if is_optional(t):
        non_optional_args: tuple[type] = tuple(
            a for a in typing.get_args(t) if a is not types.NoneType
        )
        if len(non_optional_args) == 1:
            return non_optional_args[0]
        return typing.Union[non_optional_args]  # type: ignore

    return t


def get_option(t: naff.OptionTypes | type):
    t = filter_extras(t)

    if isinstance(t, naff.OptionTypes):
        return t

    if t == str:
        return naff.OptionTypes.STRING
    if t == int:
        return naff.OptionTypes.INTEGER
    if t == bool:
        return naff.OptionTypes.BOOLEAN
    if issubclass_failsafe(t, (naff.BaseUser, naff.Member)) or t in USER_UNIONS:
        return naff.OptionTypes.USER
    if issubclass_failsafe(t, naff.BaseChannel):
        return naff.OptionTypes.CHANNEL
    if t == naff.Role:
        return naff.OptionTypes.ROLE
    if t == float:
        return naff.OptionTypes.NUMBER
    if t == naff.Attachment:
        return naff.OptionTypes.ATTACHMENT
    if t in MENTIONABLE_UNIONS:
        return naff.OptionTypes.MENTIONABLE

    if is_union(t):
        args = typing.get_args(t)
        if all(issubclass_failsafe(a, naff.BaseChannel) for a in args):
            return naff.OptionTypes.CHANNEL

    raise ValueError("Invalid type provided.")


def resolve_channel_types(anno: typing.Any):
    channel_types = []

    anno = filter_extras(anno)

    if isinstance(anno, naff.OptionTypes):
        return None

    if issubclass_failsafe(anno, naff.BaseChannel) and (
        chan_type := REVERSE_CHANNEL_MAPPING.get(anno)
    ):
        channel_types = [chan_type]
    elif is_union(anno):
        for arg in typing.get_args(anno):
            if chan_type := REVERSE_CHANNEL_MAPPING.get(arg):
                channel_types.append(chan_type)

    return channel_types or None
